%******************************************************************************;
%* SAS macro NumCodeC vs. 0.0.2 , 27 Apr 2011, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%* STILL UNDER CONSTRUCTION, NOT FOR VALIDATION, USE WITH CAUTION AT OWN RISK *;
%******************************************************************************;
%*                                                                            *;
   %MACRO   NumCodeC   /* Numeric code character values for ordered tabulation*/
/* -------------------------------------------------------------------------- */
/* ---------- A l w a y s   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  Data    = _LAST_    /* Input data set specification                      R */
,  Out     =           /* Output data set specification (may be InputData)  R */
                       /*    (if empty then same as input data set spec.)     */
,  OverWrit= No        /* specify Yes to overwrite an existing Out dataset  R */
,  Variable=           /* character variable name to recode to numeric var  R */
                       /*    (recoded numeric variable may have same name,    */
                       /*    in that case the original variable gets lost,    */
                       /*    see also NumCoded variable name argument below)  */ 
,  nFmtName=           /* numeric format name with original, character      R */
                       /*    values as formatted values (in specified order)  */
,  FmtOrder=Alpha      /* Alpha (original alphabetic) / Freq / InFormat     R */
                       /*    These keywords may be extended with: Desc        */
/*                                                                            */
/* ------ F r e q u e n t l y   s p e c i f i e d   a r g u m e n t s ------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  NumCoded=           /* numerically recoded variable name (if different)  O */
                       /*   if empty then original variable name              */
,  InFormat=           /* existing character informat name to order values  O */
/*                                                                            */
/* ---- O c c a s i o n a l l y   s p e c i f i e d   a r g u m e n t s ----- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Where   = 1         /* complete where clause, parentheses not necessary  O */
,  Version = 0.0.2     /* forced version specification for version control  O */
/*                                                                            */
/* ----------- N e v e r   s p e c i f i e d   a r g u m e n t s ------------ */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the program's use only, don't use it O */
)/ DES     =              'Prepare char values for ordered tables'  /*Store*/  ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : NumCodeC.SAS / NumCodeC.002                                   *;
%*                                                                            *;
%* Purpose    : Automatically recode and order character variable values to   *;
%*              numeric ones with original values as formats for tabulation   *;
%*                                                                            *;
%* Programmer : Jim Groeneveld, OCS Biometric Support, `s Hertogenbosch, NL   *;
%* Date       : 21 - 27 Apr 2011                                              *;
%* Version    : 0.0.2 / useful but still very much under construction         *;
%*                                                                            *;
%* SAS version: 9.1/9.2 under MS Windows XP, Professional Edition / EG 4.3/XP *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Summary:                                                                   *;
%* ========                                                                   *;
%*                                                                            *;
%* Remarks:                                                                   *;
%* ========                                                                   *;
%*                                                                            *;
%* Syntax:                                                                    *;
%* =======                                                                    *;
%*                                                                            *;
%* Examples:                                                                  *;
%* =========                                                                  *;
%*                                                                            *;
%* Output:                                                                    *;
%* =======                                                                    *;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: -                                                              *;
%* = attached: -                                                              *;
%* = external: %N0Cases, %ChkExist, %_Delete_                                 *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of NumCodeC:                                                       *;
%* ====================                                                       *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0.0    22 Apr 2011  first working release (started on 21 Apr 2011)       *;
%* 0.0.1    22 Apr 2011  added InFormat argument and FmtOrder=InFormat        *;
%* 0.0.2    22 Apr 2011  support for extraneous values of &Variable not       *;
%*               to        contained in &InFormat specification: add at end,  *;
%*          27 Apr 2011  many improvements and bug fixes with missing data    *;
%*                                                                            *;
%* Future of NumCodeC:                                                        *;
%* ===================                                                        *;
%* - enable formatted text values over 40 characters !!! (not possible???)    *;
%* - automatic and/or split wrapping of (too) long formatted text values      *;
%*                                                                            *;
%******************************************************************************;

 %LOCAL    /* Define internal macro variables explicitely local */
           /* GENERAL */
  MacName  /* This macros external name: NumCodeC */
  MacVs    /* Version number major.minor.patchSub of macro: 0.0.2 */
  MacDate  /* Date of this version of macro: 27 Apr 2011 */
  ErrCount /* number of erroneously specified arguments */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  Timing   /* Start time or run time */
  ;

  %LET Timing   = %SYSFUNC(DATETIME());
  %LET MacName  = NumCodeC;
  %LET MacVs    = 0.0.2;
  %LET MacDate  = 27 Apr 2011;
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R =, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G =, %STR( )));

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %GOTO Finish;

  %IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
    OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro header in LOG file;
%*_________________________;

  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. &MacVs, by Jim Groeneveld, &MacDate  |;
  %PUT | === Conversion of character values to numeric categories === |;
  %PUT | === with format with generated original formatted values === |;
  %PUT |   Version 0.x.y: NOT yet VALIDATED, USE at YOUR own RISK!    |;
  %PUT |______________________________________________________________|;
  %PUT;

%* Feedback of specified arguments;
  %PUT &MacName: Review of argument specifications (incl. defaults);
  %PUT &MacName: Data     = &Data;
  %PUT &MacName: Out      = &Out;
  %PUT &MacName: Overwrit = &Overwrit;
  %PUT &MacName: Variable = &Variable;
  %PUT &MacName: nFmtName = &nFmtName;
  %PUT &MacName: FmtOrder = &FmtOrder;
  %PUT &MacName: NumCoded = &NumCoded;
  %PUT &MacName: InFormat = &InFormat;
  %PUT &MacName: Where    = &Where;
  %PUT &MacName: Version  = &Version;
  %PUT &MacName: Debug    = &Debug;
  %PUT &MacName: ForMyUse = &ForMyUse;

  %IF (&Version NE AND &Version NE &MacVs) %THEN %DO;
    %PUT *** &MacName *** &ERR Specified version &Version does not match &MacVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Exit;
  %END;

%* Process keyword _LAST_ with Data;
  %IF (%UPCASE(&Data) EQ _LAST_) %THEN %LET Data = &SYSLAST;

 %LOCAL    /* Define internal macro variables explicitely local */
           /* SPECIFIC */
  LibName  /* Deduced LIBNAME from Data macro variable */
  A E F I J K L R /* multi-purpose loop counters, index variables */
  WorkData /* local work dataset name of Data after Where */
  Sequence /* Logical record sequence number name in &WorkData */
  NofCases /* Number of logical observations in dataset &Data*/
  FreqData /* output dataset name of PROC FREQ with values and counts */
  NumValue /* variable name of numerical order values for &Variable values */
  Ordering /* storage of FmtOrder keyword (not Desc) */
  Desc     /* storage of keyword DESCENDING with FmtOrder=order Desc */
  cFmtData /* output dataset name from forced ordered character format */
  NofCases /* Number of records in temporary datasets for checking on empty dataset */
  ;

  %LET WorkData = __Work__;
  %LET Sequence = _SeqNum_;
  %LET FreqData = __Freq__;
  %LET NumValue = _CatNum_;
  %LET cFmtData = __cFmt__;

%PUT ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT *** Check arguments for legal contents and discrepancies, calling e_r_r_o_r_s ***;
%PUT _________________________________________________________________________________;

  %LET ErrCount = 0;
%* Data argument;
  %IF (%LENGTH(&Data) EQ 0) %THEN %DO;
    %PUT *** &MacName *** &ERR Dataset not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (%SYSFUNC(EXIST(&Data)) EQ 0) %THEN %DO;
    %PUT *** &MacName *** &ERR Dataset &Data does not exist,;
    %PUT                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
%* See also SAS Guide to macro processing, vs. 6, 2nd ed. p. 265: macro Exist;
  %END;

%* Out argument;
  %IF (%LENGTH(&Out) EQ 0) %THEN %DO; %* make Out spec identical to Data spec;
    %LET Out = &Data;
  %END;

  %IF ( %FirstCh(&Overwrit) NE Y AND %FirstCh(&Overwrit) NE N) %THEN %DO;
    %PUT *** &MacName *** &ERR Overwrit argument not specified or illegal: &Overwrit,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF ( %SYSFUNC(EXIST(&Out)) AND %FirstCh(&OverWrit) NE Y ) %THEN %DO;
    %PUT *** &MacName *** &ERR Dataset Out=&Out exists and OverWrit is not Yes,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

/* check Variable's existence */
  %IF (%ChkExist(Data=&Data, VarList=&Variable)) %THEN %GOTO Aborting; %* 0 = all exist;

/* check NumCoded's non-existence if not equal to Variable */
  %IF (NOT %LENGTH(&NumCoded)) %THEN %LET NumCoded = &Variable; %* make equal;
  %IF (%UPCASE(&Variable) NE %UPCASE(&NumCoded)) %THEN %DO; 
    %IF (%CHkExist(Data=&Data, VarList=&NumCoded, CheckVar=Absence)) %THEN
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

%* Check nFmtName for numerical type;
  %IF (NOT %LENGTH(&nFmtName)) %THEN %DO;
    %PUT *** &MacName *** &ERR nFmtName=&nFmtName not specified, is required;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (%FirstCh(&nFmtName) EQ $) %THEN %DO;
    %PUT *** &MacName *** &ERR nFmtName=&nFmtName is a character format name;
    %PUT .                     and must be a numerical format name (without $);
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

%* Check InFormat for Character type;
  %IF (%FirstCh(&InFormat) NE $ AND %LENGTH(&InFormat)) %THEN %DO;
    %PUT *** &MacName *** &ERR InFormat=&InFormat is a numerical format name;
    %PUT .                     and must be a character format name (with $);
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

%* check FmtOrder for specific values;
%* Firstly separate order from optional DESCending value;
  %LET Desc = ; %* empty;
  %LET Ordering = ; %* empty;
  %IF (%UPCASE(%SCAN(&FmtOrder,2,%STR( ))) EQ DESC) %THEN %LET Desc = DESCENDING;
  %LET Ordering = %UPCASE(%SCAN(&FmtOrder,1,%STR( )));
  %IF (&Ordering EQ ALPHA OR &Ordering EQ FREQ 
    OR &Ordering EQ INFORMAT) %THEN /* nothing */;
  %ELSE %DO;
    %PUT *** &MacName *** &ERR Illegal FmtOrder=&FmtOrder value(s) specified,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF (&Ordering EQ INFORMAT AND NOT %LENGTH(&InFormat)) %THEn %DO;
    %PUT *** &MacName *** &ERR FmtOrder=&FmtOrder implies a character informat,;
    %PUT .                     but that format has not been specified (Informat=);
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

%* Remove possible trailing dot from nFmtName and InFormat;
  %LET nFmtName = %SCAN(&nFmtName, 1, .);
  %IF (%LENGTH(&InFormat)) %THEN %LET InFormat = %SCAN(&InFormat, 1, .);

%* Where - no check, except for being not empty (1);
  %IF (NOT %LENGTH(&Where)) %THEN %LET Where = 1;

%* Version - no check;
%* Debug - no check:
%* ForMyUse - no check;


  %IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;
%* -------------------------------end-of-checks------------------------------- ;
%**GOTO Finish; %*** debug;

%PUT ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT *** Various initialisations and preparations ***;
%PUT ________________________________________________;

%PUT *** Prepare work dataset and assign logical sequence numbers ***;
  DATA &WorkData; SET &Data (WHERE=(&Where)); &Sequence+1; RUN;

%PUT *** Process empty dataset due to &Where ***;
  %N0cases (Data=&WorkData, Message="Dataset &Data~where (&Where) is empty - no recode!",
            Returned=NofCases);
  %IF (&NofCases LE 0) %THEN %GOTO Exit;


%*======;
%* Core ; %*==================================================;
%*======;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * First of all a PROC FREQ to obtain the existing character values and their counts ;
%PUT *___________________________________________________________________________________;

  PROC FREQ DATA=&WorkData NOPRINT;
    TABLES &Variable / OUT=&FreqData (KEEP=&Variable Count);
  RUN;

  %IF (&Debug GE 10) %THEN %DO;
    data sas__out.freqdata; set &freqdata; run; * debug;
  %END;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * Order values according to desired sorted order   ;
%PUT * Alpha / Freq / Alpha Desc / Freq Desc / Informat ;
%PUT *__________________________________________________;

  %IF (%UPCASE(&Ordering) EQ ALPHA) %THEN %DO;
    PROC SORT DATA=&FreqData; BY &Desc &Variable; RUN;
  %END;

  %ELSE %IF (%UPCASE(&Ordering) EQ FREQ) %THEN %DO;
    PROC SORT DATA=&FreqData; BY &Desc Count; RUN;
  %END;

  %ELSE %IF (%UPCASE(&Ordering) EQ INFORMAT) %THEN %DO;
%PUT * - - - - - - - - - - - - - - - - - - - - - - - - - ;
%PUT * make dataset with original character values and   ;
%PUT * assigned order from user defined character format ;
%PUT * - - - - - - - - - - - - - - - - - - - - - - - - - ;
    PROC FORMAT /*MAXLABLEN=255 MAXSELEN=255 -- no effect */
                CNTLOUT=&cFmtData (KEEP=Start Label FmtName
                RENAME=(Start=&Variable)
                WHERE=(UPCASE(FmtName) EQ UPCASE(SUBSTR("&InFormat", 2))));

    %IF (&Debug GE 10) %THEN %DO;
      data sas__out.cFmtData1; set &cFmtData; run; * debug;
    %END;

%PUT * renaming Label into &NumValue yields a character variable ;
%PUT * so a type conversion has to take place instead ;
    DATA &cFmtData (DROP=Label);
      SET &cFmtData (DROP=FmtName);
      &NumValue = INPUT (Label, BEST12.); * also all 28 missing values;
    RUN;

    %IF (&Debug GE 10) %THEN %DO;
      data sas__out.cFmtData2; set &cFmtData; run; * debug;
    %END;

%PUT * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%PUT * Both &FreqData and &cFmtData ordered by original &Variable by default.;
%PUT * Merge &FreqData and &cFmtData by &Variable in order to add extraneous,;
%PUT * &Variable values not specified in &InFormat to the end of &nFmtName.  ;
%PUT * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

    DATA &FreqData;
      MERGE &FreqData &cFmtData (IN=IncFmt);
        BY &Variable;
* Set extraneous values of &Variable=Label to high value 99999;
      IF (/*MISSING(&NumValue)*/ NOT IncFmt) THEN DO;
        &NumValue = 99999; * (specified missing value remains missing);
* Format occurring empty data as missing, unless explicitly different;
        IF (MISSING(&Variable)) THEN &NumValue = .;
      END;
* (Leave explicitly specified empty values as explicitly formatted);
    RUN;

    %IF (&Debug GE 10) %THEN %DO;
      data sas__out.FreqData2; set &FreqData; run; * debug;
    %END;

%PUT * sort that dataset according to the specified order;
    PROC SORT DATA=&FreqData; BY &Desc &NumValue; RUN;
  %END; %* InFormat;

  %IF (&Debug GE 10) %THEN %DO;
    data sas__out.FreqData3; set &FreqData; run; * debug;
  %END;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * Assign order values from 1 in &FreqData: &NumCoded ;
%PUT * but initially store these in variable &NumValue to ;
%PUT * be used for generating a corresponding character   ;
%PUT * format. Also insert FmtName variable as &nFmtName. ;
%PUT *____________________________________________________;

  DATA &FreqData (DROP=Ordering);
    SET &FreqData;
    IF (NOT MISSING(&NumValue)) THEN DO;
      Ordering + 1; * temporary variable;
      &NumValue = Ordering; * (not influenced by missing);
    END;
    FmtName = "&nFmtName";
    Type = 'N';
* @@@@@@@@@ try to wrap (too) long formatted values (&Variable) here;
  RUN;

  %IF (&Debug GE 10) %THEN %DO;
    data sas__out.FreqData4; set &FreqData; run; * debug;
  %END;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * Create numeric format from &FreqData for &NumValue ;
%PUT *____________________________________________________;

  PROC FORMAT CNTLIN=&FreqData (RENAME=(&NumValue=Start &Variable=Label));
  RUN;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Associate format &nFmtName with &NumValue values ;
* is integrated into next, final data step below.  ;
*__________________________________________________;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * Merge the NumCoded data to the original data, possibly reassigning &Variable ;
%PUT * Firstly sort both by &Variable and the result by the original &Sequence      ;
%PUT *______________________________________________________________________________;

  PROC SORT DATA=&WorkData; BY &Variable; RUN;
  PROC SORT DATA=&FreqData; BY &Variable; RUN;

  DATA &WorkData 
      %IF (%UPCASE(&Variable) EQ %UPCASE(&NumCoded)) %THEN (DROP=&Variable);
      ; * drop existing &Variable if &NumCoded is same;
    MERGE &WorkData (IN=InWork)
          &FreqData (KEEP=&Variable &NumValue);
      BY &Variable;
    IF (InWork);
    FORMAT &NumValue &nFmtName..;
  RUN;

%PUT *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%PUT * Reorder data to original order (&Sequence).;
%PUT * Rename temporary variable with category    ;
%PUT * numbers to (same) indicated variable name. ;
%PUT *____________________________________________;

  PROC SORT DATA=&WorkData (RENAME=(&NumValue=&NumCoded))
             OUT=&Out (DROP=&Sequence);
    BY &Sequence;
  RUN;

  %Interupt:
  %Aborting:
  %Exit:
  %Finish:

  %LET SYSLAST = &Data; %* Reset to last outer dataset used;
  %_Delete_ (&WorkData &FreqData &cFmtData)

%* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* *** Process start time and stop time into run time *** ;
%* _______________________________________________________;

  %IF ( %UPCASE(&ForMyUse) NE RECURSIVE ) %THEN
  %DO;
    %LET Timing   = %SYSEVALF ( %SYSFUNC(DATETIME()) - &Timing );
%* Elapsed time;
    %PUT ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
    %PUT *** &MacName *** NOTE: The MACRO &MacName used &Timing seconds.;
  %END;

  %IF ( &ErrCount GT 0 ) %THEN
  %DO;
    DATA _NULL_;
      ABORT /*ABEND*/;
    RUN;
  %END;

  %PUT ---------------------------------------------------------------;
  %PUT *** &MacName *** NOTE: End of macro &MacName ================= ;
  %PUT _______________________________________________________________;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%MEND NumCodeC;
%*___________________________________________________________________;

